"""
.. module:: util.py

"""

import urllib
import re
import codecs
from twisted.internet import defer
from twisted.internet import reactor
from zope.interface import implements
from twisted.web import client
from twisted.web.http import PotentialDataLoss
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer
from twisted.internet.protocol import Protocol
from twisted.web._newclient import ResponseDone
from StringIO import StringIO

DEBUG=False

# XXX pyOpenSSL is basically fucked in a whole host of distro's
#     thanks to SSLv2 being summarily removed from libssl. 
#     For the time being, to hell with SSL, it breaks everywhere
#from twisted.internet.ssl import ClientContextFactory

#class WebClientContextFactory(ClientContextFactory):
#    def getContext(self, hostname, port):
#        return ClientContextFactory.getContext(self)


class Agent(client.Agent):
    """
    client.Agent does not provide the ability to set a timeout.
    This is an Agent with a timeout.
    """

    def __init__(self, reactor,
                       contextFactory=None,
                       timeout=240):
        self._reactor = reactor
        self._contextFactory = contextFactory
        self.timeout = timeout


    def _connect(self, scheme, host, port):
        """ connect method with a timeout. """

        cc = client.ClientCreator(self._reactor, self._protocol)
        if scheme == 'http':
            d = cc.connectTCP(host, port, timeout=self.timeout)
        elif scheme == 'https':
            raise Exception('HTTPS not supported') 
            #d = cc.connectSSL(host, port, self._wrapContextFactory(host, port),
            #                  timeout=self.timeout)
        else:
            d = defer.fail(SchemeNotSupported(
                    "Unsupported scheme: %r" % (scheme,)))
        return d

class StringProducer(object):
    """
    Body producer for t.w.c.Agent
    """
    implements(IBodyProducer)

    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        return defer.maybeDeferred(consumer.write, self.body)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


class ResponseReceiver(Protocol):
    """
    Assembles HTTP response from return stream.
    """

    def __init__(self, deferred):
        self.writer = codecs.getwriter("utf_8")(StringIO())
        self.deferred = deferred

    def dataReceived(self, bytes):
        self.writer.write(bytes)

    def connectionLost(self, reason):
        if reason.check(ResponseDone) or reason.check(PotentialDataLoss):
            self.deferred.callback(self.writer.getvalue())
        else:
            self.deferred.errback(reason)

def getPageWithHeaders(contextFactory=None,
                      scheme='http', host='127.0.0.1', port=80, path='/',
                      timeout=240, *args, **kwargs):
    """Download a web page as a string.

    :returns: twisted.web.client.Agent deferred.

    Download a page. Return a deferred, which will callback with a
    page (as a string) or errback with a description of the error.

    See t.w.c.Agent to see what extra args can be passed.
    """

    def cb_recv_resp(response):
        d_resp_recvd = defer.Deferred()
        if response.length:
            response.deliverBody(ResponseReceiver(d_resp_recvd))
            return d_resp_recvd.addCallback(cb_process_resp, response)
        else:
            return cb_process_resp("", response)

    def cb_process_resp(body, response):
        _headers = {"http_code": response.code}

        for header in response.headers.getAllRawHeaders():
            _headers[header[0].lower()] = header[1][0]

        return _headers, body

    url = str("http://%s:%d%s" % (host, port, path))

    if DEBUG:
        print "==========================="
        print ">>uri: %s" % url
        print ">>method: %s" % kwargs.get('method', None)
        print ">>headers: %s" % kwargs.get('headers', None)
        print ">>cookies: %s" % kwargs.get('cookies', None)
        print ">>postdata: %s" % kwargs.get('postdata', None)

    if not "headers" in kwargs:
        kwargs["headers"] = {}
    else:
        for header in kwargs["headers"]:
            kwargs["headers"][header] = [kwargs["headers"][header]]

    if not "method" in kwargs:
        kwargs["method"] == "GET"

    if "postdata" in kwargs:
        body = StringProducer(kwargs["postdata"])
    else:
        body = None

    d = Agent(reactor, timeout=timeout).request(kwargs["method"],
                                      url,
                                      Headers(kwargs["headers"]),
                                      body
                                  )

    d.addCallback(cb_recv_resp)
    return d

def flatten_js(js_func):
    """Flatten a JavaScript function into a single line.

    :returns: String.

    Take a JavaScript function and strip out all of the
    newlines so that it is suitable for passing into a
    Riak map/reduce phase.
    """

    return "".join(js_func.split("\n"))

def get_value(key, array, default_value):
    """
    Overkill for array.get(key, default_value).
    Likely due to porting from another language.
    """
    if (key in array):
        return array[key]
    else:
        return default_value

def build_rest_path(client,
                    bucket=None, key=None, spec=None, params=None,
                    prefix=None):
    """
    Given a RiakClient, RiakBucket, Key, LinkSpec, and Params,
    construct and return a URL.
    """
    # Build 'http://hostname:port/prefix/bucket'
    path = ''
    path += '/' + (prefix or client._prefix)

    # Add '.../bucket'
    if bucket:
        path += '/' + urllib.quote_plus(bucket._name)

    # Add '.../key'
    if bucket and key:
        path += '/' + urllib.quote_plus(key)

    # Add query parameters.
    if (params != None):
        s = ''
        for key in params.keys():
            if (s != ''):
                s += '&'
            s += (urllib.quote_plus(key) + '=' +
                  urllib.quote_plus(str(params[key])))
        path += '?' + s

    # Return.
    return client._host, client._port, path

def http_request_deferred(method, host, port, path,
                          headers=None, obj=''):
    """
    Given a Method, URL, Headers, and Body, perform an HTTP request,
    and return deferred.
    :returns: deferred
    """
    if headers == None:
        headers = {}

    return getPageWithHeaders(contextFactory=None,
                              scheme='http',
                              host=host,
                              method=method,
                              port=port,
                              path=path,
                              headers=headers,
                              postdata=obj,
                              timeout=20)

def build_headers(headers):
    """
    Build up the header string.
    """
    headers1 = []
    for key in headers.keys():
        headers1.append('%s: %s' % (key, headers[key]))
    return headers1

def parse_http_headers(headers):
    """
    Parse an HTTP Header string into dictionary of response headers.
    """
    result = {}
    fields = headers.split("\n")
    for field in fields:
        matches = re.match("([^:]+):(.+)", field)
        if (matches == None):
            continue
        key = matches.group(1).lower()
        value = matches.group(2).strip()
        if (key in result.keys()):
            if  isinstance(result[key], list):
                result[key].append(value)
            else:
                result[key] = [result[key]].append(value)
        else:
            result[key] = value
    return result





"""
.. module:: bucket.py

Riak bucket objects

"""

import urllib
import json
from twisted.internet import defer

from riakasaurus import riak_object, util


class RiakBucket(object):
    """
    The RiakBucket object allows you to access and change information
    about a Riak bucket, and provides methods to create or retrieve
    objects within the bucket.
    """

    SEARCH_PRECOMMIT_HOOK = {"mod": "riak_search_kv_hook", "fun": "precommit"}

    def __init__(self, client, name):
        """
        Initialize RiakBucket
        """
        self._client = client
        self._name = name
        self._r = None
        self._w = None
        self._dw = None
        return None

    def get_name(self):
        """
        Get the bucket name.
        .. todo:: remove accessors
        """
        return self._name

    def get_r(self, r=None):
        """
        Get the R-value for this bucket, if it is set, otherwise return
        the R-value for the client.
        :rtype: return integer
        .. todo:: make into property
        """
        if (r != None):
            return r
        if (self._r != None):
            return self._r
        return self._client.get_r()

    def set_r(self, r):
        """
        Set the R-value for this bucket. get(...) and get_binary(...)
        operations that do not specify an R-value will use this value.
        @param integer r - The new R-value.
        @return self
        .. todo:: convert to property
        """
        self._r = r
        return self

    def get_w(self, w):
        """
        Get the W-value for this bucket, if it is set, otherwise return
        the W-value for the client.
        @return integer
        .. todo:: convert to property
        """
        if (w != None):
            return w
        if (self._w != None):
            return self._w
        return self._client.get_w()

    def set_w(self, w):
        """
        Set the W-value for this bucket. See set_r(...) for more information.
        @param integer w - The new W-value.
        @return self
        .. todo:: convert to property
        """
        self._w = w
        return self

    def get_dw(self, dw):
        """
        Get the DW-value for this bucket, if it is set, otherwise return
        the DW-value for the client.
        @return integer
        .. todo:: convert to property
        """
        if (dw != None):
            return dw
        if (self._dw != None):
            return self._dw
        return self._client.get_dw()

    def set_dw(self, dw):
        """
        Set the DW-value for this bucket. See set_r(...) for more information.
        @param integer dw - The new DW-value
        @return self
        .. todo:: convert to property
        """
        self._dw = dw
        return self

    def new(self, key, data=None):
        """
        Create a new Riak object that will be stored as JSON.

        :param key: Name of the key.
        :param data: The data to store.
        :returns: riak_object.RiakObject
        """
        obj = riak_object.RiakObject(self._client, self, key)
        obj.set_data(data)
        obj.set_content_type('text/json')
        obj._jsonize = True
        return obj

    def new_binary(self, key, data, content_type='text/json'):
        """
        Create a new Riak object that will be stored as plain text/binary.

        :param key: Name of the key.
        :param data: The data to store.
        :param content_type: The content type of the object.
        :returns: riak_object.RiakObject
        """
        obj = riak_object.RiakObject(self._client, self, key)
        obj.set_data(data)
        obj.set_content_type(content_type)
        obj._jsonize = False
        return obj

    @defer.inlineCallbacks
    def get(self, key, r=None):
        """
        Retrieve a JSON-encoded object from Riak.

        :param key: Name of the key.
        :param r: R-Value of this request (defaults to bucket's R)
        :returns: riak_object.RiakObject -- via deferred
        """
        obj = riak_object.RiakObject(self._client, self, key)
        obj._jsonize = True
        r = self.get_r(r)
        result = yield obj.reload(r)
        defer.returnValue(result)

    @defer.inlineCallbacks
    def get_binary(self, key, r=None):
        """
        Retrieve a binary/string object from Riak.

        :param key: Name of the key.
        :param r: R-Value of this request (defaults to bucket's R)
        :returns: riak_object.RiakObject -- via deferred
        """
        obj = riak_object.RiakObject(self._client, self, key)
        obj._jsonize = False
        r = self.get_r(r)
        result = yield obj.reload(r)
        defer.returnValue(result)

    @defer.inlineCallbacks
    def set_n_val(self, nval):
        """
        Set the N-value for this bucket, which is the number of replicas
        that will be written of each object in the bucket. Set this once
        before you write any data to the bucket, and never change it
        again, otherwise unpredictable things could happen. This should
        only be used if you know what you are doing.
        @param integer nval - The new N-Val.
        .. todo:: Given the danger, some way to first check for existence?
        """
        result = yield self.set_property('n_val', nval)
        defer.returnValue(result)

    @defer.inlineCallbacks
    def get_n_val(self):
        """
        Retrieve the N-value for this bucket.
        @return integer
        .. todo:: what happens if you ask for n_val before ever writing?
        """
        result = yield self.get_property('n_val')
        defer.returnValue(result)

    @defer.inlineCallbacks
    def set_allow_multiples(self, the_bool):
        """
        If set to True, then writes with conflicting data are stored
        and returned to the client. This situation can be detected by
        calling has_siblings() and get_siblings(). This should only be used
        if you know what you are doing.

        :param the_bool: True to store and return conflicting writes.
        :returns: deferred
        """
        result = yield self.set_property('allow_mult', the_bool)

        defer.returnValue(result)

    @defer.inlineCallbacks
    def get_allow_multiples(self):
        """
        Retrieve the 'allow multiples' setting.

        :returns: Boolean -- via deferred
        """
        result = yield self.get_property('allow_mult')

        defer.returnValue(result)

    @defer.inlineCallbacks
    def set_property(self, key, value):
        """
        Set a bucket property. This should only be used if you know what
        you are doing.

        :param key: property to set.
        :param value: property value to set.
        :returns: deferred result
        """
        result = yield self.set_properties({key: value})

        defer.returnValue(result)

    @defer.inlineCallbacks
    def get_property(self, key):
        """
        Retrieve a bucket property.

        :param key: property to retrieve
        :returns: property value -- as deferred
        """

        props = yield self.get_properties()

        if (key in props.keys()):
            defer.returnValue(props[key])
        else:
            defer.returnValue(None)

    @defer.inlineCallbacks
    def search_enabled(self):
        """
        Returns True if the search precommit hook is enabled for this bucket.
        """
        props = yield self.get_property("precommit") or []
        defer.returnValue(self.SEARCH_PRECOMMIT_HOOK in props)

    @defer.inlineCallbacks
    def enable_search(self):
        precommit_hooks = yield self.get_property("precommit") or []

        if self.SEARCH_PRECOMMIT_HOOK not in precommit_hooks:
            yield self.set_properties({"precommit":
                precommit_hooks + [self.SEARCH_PRECOMMIT_HOOK]})

        defer.returnValue(True)

    @defer.inlineCallbacks
    def disable_search(self):
        precommit_hooks = yield self.get_property("precommit") or []

        if self.SEARCH_PRECOMMIT_HOOK in precommit_hooks:
            precommit_hooks.remove(self.SEARCH_PRECOMMIT_HOOK)
            yield self.set_properties({"precommit": precommit_hooks})

        defer.returnValue(True)

    def search(self, query, **params):
        return self._client.solr().search(self._name, query, **params)

    @defer.inlineCallbacks
    def set_properties(self, props):
        """
        Set multiple bucket properties in one call. This should only be
        used if you know what you are doing.

        :param props: a dictionary of keys and values to store.
        :returns: deferred
        """
        host, port, url = util.build_rest_path(self._client, self)
        headers = {'Content-Type': 'application/json'}
        content = json.dumps({'props': props})

        #Run the request...
        response = yield util.http_request_deferred('PUT', host,
                                                         port, url,
                                                         headers, content)

        # Handle the response...
        if (response == None):
            raise Exception('Error setting bucket properties.')

        # Check the response value...
        status = response[0]['http_code']
        if (status != 204):
            raise Exception('Error setting bucket properties.')
        defer.returnValue(response)

    @defer.inlineCallbacks
    def get_properties(self):
        """
        Retrieve a dictionary of all bucket properties.

        :returns: dictionary -- via deferred
        """

        # Run the request...
        params = {'props': 'True', 'keys': 'False'}
        host, port, url = util.build_rest_path(self._client, self,
                                                    None, None, params)
        response = yield util.http_request_deferred('GET', host,
                                                         port, url)

        # Use a riak_object.RiakObject to interpret the response.
        # We are just interested in the value.
        obj = riak_object.RiakObject(self._client, self, None)
        obj._populate(response, [200])
        if (not obj.exists()):
            raise Exception('Error getting bucket properties.')

        props = obj.get_data()
        props = props['props']
        defer.returnValue(props)

    @defer.inlineCallbacks
    def list_keys(self):
        """
        Retrieve a list of all bucket keys.

        :returns: list -- via deferred
        """
        # Create the request
        params = {"keys": "stream", "props": "false"}

        host, port, url = util.build_rest_path(self._client,
                                                    self,
                                                    None,
                                                    None,
                                                    params)
        raw_response = yield util.http_request_deferred('GET', host,
                                                             port, url)

        if raw_response[0]["http_code"] != 200:
            raise Exception('Error listing keys in bucket %s.' % self._name)

        # Hacky method to deal with the concatenated response
        # we get due to the chunked encoding method.
        parts = raw_response[1].split('{"keys":[')
        keys = []

        for part in parts[1:]:
            temp_keys = json.loads('{"keys":[' + part)["keys"]
            keys = keys + temp_keys

        defer.returnValue([urllib.unquote(x) for x in keys])

    def get_keys(self):
        return self.list_keys()

    @defer.inlineCallbacks
    def purge_keys(self):
        """
        Purge all keys from the bucket.

        :returns: None

        This is a convenience function that lists all of the keys
        in the bucket and then deletes them.

        NB: This is a VERY resource-intensive operation, and is
            IRREVERSIBLE. Be careful.
        """

        # Get the current key list
        keys = yield self.list_keys()

        # Major key-killing action
        for key in keys:
            obj = yield self.get_binary(key)
            yield obj.delete()



"""
.. module:: mapreduce.py
"""

import json
from twisted.internet import defer

from riakasaurus import util, riak_link


class RiakMapReduce(object):
    """
    The RiakMapReduce object allows you to build up and run a
    map/reduce operation on Riak.
    """

    def __init__(self, client):
        """
        Initialize a Map/Reduce object.
        :param client: a RiakClient object.
        :returns: Nothing
        """
        self._client = client
        self._phases = []
        self._inputs = []
        self._input_mode = None
        self._mapreduce = None
        return None

    def search(self, bucket, query):
        """
        Perform a map-reduce operation using Riak Search
        :param bucket: Name of the bucket with Riak Search enabled
        :param query: Riak Search query
        """
        self._input_mode = 'query'
        self._inputs = {'module':'riak_search',
                       'function':'mapred_search',
                       'arg':[bucket, query]}
        return self

    def set_mapreduce(self, mapreduce):
        """
        Manually set a map/reduce query without assembling it.
        """
        self._mapreduce = mapreduce
        return self

    def index(self, bucket, index, startkey, endkey = None):
        """
        Begin a map/reduce operation using a Secondary Index
        query.
        @param bucket - The bucket over which to perform the search.
        @param query - The search query.
        """
        self._input_mode = 'query'

        if endkey == None:
            self._inputs = {'bucket': bucket,
                            'index':index,
                            'key':startkey }
        else:
            self._inputs = {'bucket':bucket,
                            'index':index,
                            'start':startkey,
                            'end':endkey }
        return self

    def add(self, arg1, arg2=None, arg3=None):
        """
        Add inputs to a map/reduce operation. This method takes three
        different forms, depending on the provided inputs. You can
        specify either a RiakObject, a string bucket name, or a bucket,
        key, and additional arg.
        :param arg1: RiakObject or Bucket or Key Filter
        :param arg2: Key or blank
        :param arg3: Arg or blank
        :returns: RiakMapReduce

        Key filters are dictionaries of the form:

            {"bucket": "<bucket_name>",
             "key_filters": [ ["<operation>", "<arg1>", "<argN>"],
                             ]}
        """
        # avoid circular import
        from riakasaurus import riak_object
        if (arg2 == None) and (arg3 == None):
            if isinstance(arg1, riak_object.RiakObject):
                return self.add_object(arg1)
            elif isinstance(arg1, dict):
                return self.add_key_filter(arg1)
            else:
                return self.add_bucket(arg1)
        else:
            return self.add_bucket_key_data(arg1, arg2, arg3)

    def add_object(self, obj):
        """Set bucket key data with object."""
        return self.add_bucket_key_data(obj._bucket._name, obj._key, None)

    def add_bucket_key_data(self, bucket, key, data):
        """Add data to the bucket."""
        if self._input_mode == 'bucket' or \
           self._input_mode == 'key_filter':
            raise Exception(('Already added a %s, ' % self._input_mode) +
                            'can\'t add an object.')
        else:
            self._inputs.append([bucket, key, data])
            return self

    def add_bucket(self, bucket):
        """Set bucket to work with."""
        self._input_mode = 'bucket'
        self._inputs = bucket
        return self

    def add_key_filter(self, key_filter):
        """Set up key filter to use."""
        if self._input_mode == 'bucket':
            raise Exception('Already added a bucket can\'t add a key filter.')
        else:
            self._input_mode = "key_filter"
            self._inputs = key_filter
            return self

    def link(self, bucket='_', tag='_', keep=False):
        """
        Add a link phase to the map/reduce operation.
        :param bucket: Bucket name (default '_' means all buckets)
        :param tag: Tag (default '_' means all tags)
        :param keep: Boolean flag to keep results from this stage.
        :returns: self
        """
        self._phases.append(RiakLinkPhase(bucket, tag, keep))
        return self

    def map(self, function, options=None):
        """
        Add a map phase to the map/reduce operation.
        :param function: Function to run
        :param options: Optional dictionary.
        :returns: self

        The function can be either a named Javascript function:
            'Riak.mapValues'
        an anonymous JavaScript function:
            'function(...)  ... )'
        or a list:
            ['erlang_module', 'function']

        The options is an associative array containing:
            'language',
            'keep' flag, and/or
            'arg'

        .. todo:: Remove get_value and use {}.get()
        """
        if options == None:
            options = []

        if isinstance(function, list):
            language = 'erlang'
        else:
            language = 'javascript'

        mr = RiakMapReducePhase('map',
                                function,
                                util.get_value('language', options,
                                                     language),
                                util.get_value('keep', options,
                                                     False),
                                util.get_value('arg', options,
                                                     None))
        self._phases.append(mr)
        return self

    def reduce(self, function, options=None):
        """
        Add a reduce phase to the map/reduce operation.

        :param function: Function to run
        :param options: Optional dictionary.
        :returns: self

        The function can be either a named Javascript function:
            'Riak.mapValues'
        an anonymous JavaScript function:
            'function(...)  ... )'
        or a list:
            ['erlang_module', 'function']

        The optios is an associative array containing:
            'language',
            'keep' flag, and/or
            'arg'

        .. todo:: Remove get_value and use {}.get()
        """
        if options == None:
            options = []

        if isinstance(function, list):
            language = 'erlang'
        else:
            language = 'javascript'

        mr = RiakMapReducePhase('reduce',
                                function,
                                util.get_value('language', options,
                                                     language),
                                util.get_value('keep', options,
                                                     False),
                                util.get_value('arg', options,
                                                     None))
        self._phases.append(mr)
        return self

    @defer.inlineCallbacks
    def run(self, timeout=None):
        """
        Run the map/reduce operation. Returns a list of results, or an
        array of RiakLink objects if the last phase is a link phase.

        :param timeout: Timeout in seconds. Defaults to waiting forever.
        :returns: list of results  -- via deferred
        """

        if self._mapreduce:
            # explicit map/reduce has been set.
            content = self._mapreduce
        else:
            # Convert all phases to associative arrays. Also,
            # if none of the phases are accumulating, then set the last one to
            # accumulate.
            num_phases = len(self._phases)
            keep_flag = False
            query = []
            for i in range(num_phases):
                phase = self._phases[i]
                if (i == (num_phases - 1)) and (not keep_flag):
                    phase._keep = True
                if phase._keep:
                    keep_flag = True
                query.append(phase._to_array())

            # Construct the job, optionally set the timeout...
            job = {'inputs': self._inputs, 'query': query}
            if timeout != None:
                job['timeout'] = timeout

            content = json.dumps(job)

        # Do the request...
        host = self._client._host
        port = self._client._port
        url = "/" + self._client._mapred_prefix
        response = yield util.http_request_deferred('POST', host, port,
                                                          url, {}, content)
        if response[0]["http_code"] > 299:
            raise Exception("Error running map/reduce job. Error: " +
                            response[1])
        result = json.loads(response[1])

        # If the last phase is NOT a link phase, then return the result.
        if self._phases:
            lastIsLink = isinstance(self._phases[-1], RiakLinkPhase)
        else:
            lastIsLink = False

        if not lastIsLink:
            defer.returnValue(result)

        # Otherwise, if the last phase IS a link phase, then convert the
        # results to RiakLink objects.
        a = []
        for r in result:
            link = riak_link.RiakLink(r[0], r[1], r[2])
            link._client = self._client
            a.append(link)

        defer.returnValue(a)


class RiakMapReducePhase(object):
    """
    The RiakMapReducePhase holds information about a Map phase or
    Reduce phase in a RiakMapReduce operation.
    """

    def __init__(self, type, function, language, keep, arg):
        """
        Construct a RiakMapReducePhase object.

        :param type: 'map' or 'reduce'
        :param function: string or array
        :param language: 'javascript' or 'erlang'
        :param keep: True to return the output of this phase in results.
        :param arg: Additional value to pass into the map or reduce function.
        :returns: None
        """
        self._type = type
        self._language = language
        self._function = function
        self._keep = keep
        self._arg = arg
        return None

    def _to_array(self):
        """
        Convert the RiakMapReducePhase to an associative array. Used
        internally.
        """
        stepdef = {'keep': self._keep,
                   'language': self._language,
                   'arg': self._arg}

        if ((self._language == 'javascript') and
              isinstance(self._function, list)):
            stepdef['bucket'] = self._function[0]
            stepdef['key'] = self._function[1]
        elif ((self._language == 'javascript') and
                isinstance(self._function, str)):
            if ("{" in self._function):
                stepdef['source'] = self._function
            else:
                stepdef['name'] = self._function

        elif (self._language == 'erlang' and isinstance(self._function, list)):
            stepdef['module'] = self._function[0]
            stepdef['function'] = self._function[1]

        return {self._type: stepdef}


class RiakLinkPhase(object):
    """
    The RiakLinkPhase object holds information about a Link phase in a
    map/reduce operation.
    """

    def __init__(self, bucket, tag, keep):
        """
        Construct a RiakLinkPhase object.
        @param string bucket - The bucket name.
        @param string tag - The tag.
        @param boolean keep - True to return results of this phase.
        """
        self._bucket = bucket
        self._tag = tag
        self._keep = keep
        return None

    def _to_array(self):
        """
        Convert the RiakLinkPhase to an associative array. Used
        internally.
        """
        stepdef = {'bucket': self._bucket,
                   'tag': self._tag,
                   'keep': self._keep}
        return {'link': stepdef}




"""
.. module:: link.py

"""

import urllib


class RiakLink(object):
    """
    The RiakLink object represents a link from one Riak object to
    another.
    @package RiakLink
    """

    def __init__(self, bucket, key, tag=None):
        """
        Construct a RiakLink object.

        :param bucket: The bucket name
        :param key: The key name
        :param tag: The tag name
        :returns: None
        """
        self._bucket = bucket
        self._key = key
        self._tag = tag
        self._client = None
        return None

    def get(self, r=None):
        """
        Retrieve the RiakObject to which this link points.

        :param r: R-value to use for this call.
        :returns: RiakObject

        .. todo:: this looks wrong. r should be override of r, not _key.
        """
        return self._client.bucket(self._bucket).get(self._key, r)

    def get_binary(self, r=None):
        """
        Retrieve the RiakObject to which this link points, as a binary.

        :param r: R-value to use for this call.
        :returns: RiakObject

        .. todo:: this looks wrong. r should be override of r, not _key.
        """
        return self._client.bucket(self._bucket).get_binary(self._key, r)

    def get_bucket(self):
        """
        Get the bucket name of this link.

        :returns: bucket name, not the object

        .. todo:: remove accessor
        """
        return self._bucket

    def set_bucket(self, name):
        """
        Set the bucket name of this link.
        @param string name - The bucket name.
        @return self

        .. todo:: remove accessor
        """
        self._bucket = name
        return self

    def get_key(self):
        """
        Get the key of this link.
        @return string

        .. todo:: remove accessor
        """
        return self._key

    def set_key(self, key):
        """
        Set the key of this link.
        @param string key - The key.
        @return self
        .. todo:: remove accessor
        """
        self._key = key
        return self

    def get_tag(self):
        """
        Get the tag of this link.
        @return string
        .. todo:: convert to a property
        """
        if (self._tag == None):
            return self._bucket
        else:
            return self._tag

    def set_tag(self, tag):
        """
        Set the tag of this link.
        @param string tag - The tag.
        @return self
        .. todo:: convert to a property
        """
        self._tag = tag
        return self

    def _to_link_header(self, client):
        """
        Convert this RiakLink object to a link header string. Used internally.
        """
        link = ''
        link += '</'
        link += client._prefix + '/'
        link += urllib.quote_plus(self._bucket) + '/'
        link += urllib.quote_plus(self._key) + '>; riaktag="'
        link += urllib.quote_plus(self.get_tag()) + '"'
        return link

    def isEqual(self, link):
        """
        Return True if this link and self are equal.

        :param link: RiakLink object
        :returns: True if equal
        """
        is_equal = ((self._bucket == link._bucket) and
                     (self._key == link._key) and
                     (self.get_tag() == link.get_tag())
                   )
        return is_equal

if __name__ == "__main__":
    pass
